program roperCCD("P=roperCCD:,C=ccd1")

/*  This sequencer program works with ccd.db.  
 *  It allows EPICS control of a Roper detector.  The software talks to the roperServer,
 *  which in turn controls WinView or WinSpec through the Microsoft COM interface.
 *
 *  Mark Rivers
 *  Oct. 7, 2003
*/

/* This program must be compiled with the recursive option so that multiple copies
 * can run at once */
option +r;

%%#include <string.h>
%%#include <errlog.h>
%%#include <socket_utils.h>
%%#include <epicsTime.h>
%%#include <epicsEvent.h>
#include "epicsCCD.h"
#define MAX_COMMAND_SIZE 200

/* Define escaped C functions at end of file */
%% static int send_command(SS_ID ssId, struct UserVar *pVar, char *command);

/* Variables */
int    socket;
char   buffer[MAX_COMMAND_SIZE];
double exposureTime;
double totalTime;
double delay;
char   *blank_pos;
int    nchars;
int    save_single;
int    status;
int    validPath;
int    DetInStrIndex;
int    DetOutStrIndex;
%%epicsTimeStamp startTime, currentTime;
%%epicsEventId abortEvent;

ss main {
    state init {
        when() {
            socket = -1;
            efClear(SetTemp_mon);
            efClear(AcquireCLBK_mon);
            efClear(AcquirePOLL_mon);
            efClear(Abort_mon);
            efClear(SaveFile_mon);
            efClear(ServerName_mon);
            efClear(ServerPort_mon);
            efClear(BinX_mon);
            efClear(BinY_mon);
            efClear(ROITop_mon);
            efClear(ROIBottom_mon);
            efClear(ROILeft_mon);
            efClear(ROIRight_mon);
            efClear(Hours_mon);
            efClear(Minutes_mon);
            efClear(Seconds_mon);
            efClear(Milliseconds_mon);
            efClear(UserInStr_mon);
            efSet(ShutterStatLink_mon); /* Copy link first time through */
            efSet(PollDetState_mon); /* Poll detector first time through */
            ConnectState = STATE_DISCONNECTED; pvPut(ConnectState);
            Initialize = 1; pvPut(Initialize); /* Force initialize first time through */
            strncpy(CCDManufacturer, "Roper", sizeof(CCDManufacturer)); pvPut(CCDManufacturer);
            strncpy(CCDModel, "WinView/WinSpec", sizeof(CCDModel)); pvPut(CCDModel);
            BitDepth = 16; pvPut(BitDepth);
            DetInStrIndex = pvIndex(DetInStr);
            DetOutStrIndex = pvIndex(DetOutStr);
            %%abortEvent = epicsEventCreate(epicsEventEmpty);
        } state wait_connect
    }

    state wait_connect {
        when(delay(1) && ConnectState == STATE_DISCONNECTED) {
            /* Close any open socket *
            if (socket >= 0) close(socket);
            /* Try to open socket connection to server */
            socket = open_socket(ServerName, ServerPort);
            /* Set the Connected PV to indicate if it was successful */
            if (socket < 0) ConnectState = STATE_DISCONNECTED;
            else            ConnectState = STATE_CONNECTED;
            pvPut(ConnectState);
        } state wait_connect

        when(ConnectState == STATE_CONNECTED) {
            %%send_command(ssId, pVar, "get_controller");
            strncpy(CCDModel, buffer, sizeof(CCDModel)); pvPut(CCDModel);
        } state monitor_changes
    }

    state monitor_changes {
        when((ConnectState == STATE_DISCONNECTED) ||
             (efTestAndClear(Initialize_mon) && Initialize == 1) ||
             (efTestAndClear(ServerName_mon)) ||
             (efTestAndClear(ServerPort_mon))) {
            /* Set the PV back to 0 */
            Initialize = 0; pvPut(Initialize);
        } state wait_connect

        when(efTestAndClear(PollDetState_mon)) { 
            %%send_command(ssId, pVar, "get_state");
            DetectorState = atoi(buffer); pvPut(DetectorState);
            %%send_command(ssId, pVar, "get_binning");
            sscanf(buffer, "%d,%d", &ActualBinX, &ActualBinY);
            pvPut(ActualBinX); pvPut(ActualBinY);
            %%send_command(ssId, pVar, "get_temperature");
            sscanf(buffer, "%lf", &MeasuredTemp);
            pvPut(MeasuredTemp);
            %%send_command(ssId, pVar, "get_exposure");
            sscanf(buffer, "%lf", &ActualSeconds);
            pvPut(ActualSeconds);
            %%send_command(ssId, pVar, "get_nframes");
            sscanf(buffer, "%d", &ActualNumFrames);
            pvPut(ActualNumFrames);
            %%send_command(ssId, pVar, "get_roi");
            sscanf(buffer, "%d,%d,%d,%d", 
                   &ActualROITop,  &ActualROIBottom, 
                   &ActualROILeft, &ActualROIRight);
            pvPut(ActualROITop);
            pvPut(ActualROIBottom);
            pvPut(ActualROILeft);
            pvPut(ActualROIRight);
        } state monitor_changes

        when(efTestAndClear(ShutterStatLink_mon)) {
            /* Copy the link to the shutter PV, but strip off CP MS*/
            blank_pos = strchr(ShutterStatLink, ' ');
            if (blank_pos != 0) nchars = blank_pos-ShutterStatLink; 
            else nchars=strlen(ShutterStatLink);
            strncpy(buffer, ShutterStatLink, sizeof(buffer));
            buffer[nchars] = 0;
            pvAssign(Shutter, buffer);
            /* Process the shutter status PV so the link status will change */
            /* Need to wait a little while for link to establish before processing status record */
            epicsThreadSleep(0.1);
            ShutterStatProc = 1; pvPut(ShutterStatProc);
        } state monitor_changes

        when(efTestAndClear(OpenShutter_mon) && 
                            OpenShutter==1) {
            /* Are we using an EPICS PV for the shutter, or the MAR 165 shutter? */
            if (ShutterMode == SHUTTER_EPICS) { 
                strncpy(Shutter, OpenShutterStr, sizeof(Shutter)); pvPut(Shutter);
            } else if (ShutterMode == SHUTTER_CAMERA) {
                %%send_command(ssId, pVar, "shutter,1");
            }
            /* Set the PV back to 0 */
            OpenShutter = 0; pvPut(OpenShutter);
        } state monitor_changes

        when(efTestAndClear(CloseShutter_mon) && 
                            CloseShutter==1) {
            /* Are we using an EPICS PV for the shutter, or the MAR 165 shutter? */
            if (ShutterMode == SHUTTER_EPICS) { 
                strncpy(Shutter, CloseShutterStr, sizeof(Shutter)); pvPut(Shutter);
            } else if (ShutterMode == SHUTTER_CAMERA) {
                %%send_command(ssId, pVar, "shutter,0");
            }
            /* Set the PV back to 0 */
            CloseShutter = 0; pvPut(CloseShutter);
        } state monitor_changes

        when((efTestAndClear(BinX_mon)) ||
             (efTestAndClear(BinY_mon))) {
            /* Set binning */
            snprintf(buffer, sizeof(buffer), "set_binning,%d,%d", BinX, BinY);
            %%send_command(ssId, pVar, pVar->buffer);
        } state monitor_changes

        when((efTestAndClear(Hours_mon)) ||
             (efTestAndClear(Minutes_mon)) ||
             (efTestAndClear(Seconds_mon)) ||
             (efTestAndClear(Milliseconds_mon))) {
            /* Start acquisition */
            exposureTime = Hours*3600. + Minutes*60. + Seconds + Milliseconds/1000.;
            snprintf(buffer, sizeof(buffer), "set_exposure,%f", exposureTime);
            %%send_command(ssId, pVar, pVar->buffer);
        } state monitor_changes

        when((efTestAndClear(ROITop_mon)) ||
             (efTestAndClear(ROIBottom_mon)) ||
             (efTestAndClear(ROILeft_mon)) ||
             (efTestAndClear(ROIRight_mon))) {
            /* Set ROI */
            snprintf(buffer, sizeof(buffer), "set_roi,%d,%d,%d,%d", ROITop, ROIBottom, ROILeft, ROIRight);
            %%send_command(ssId, pVar, pVar->buffer);
        } state monitor_changes

        when(efTestAndClear(UserInStr_mon)) {
            /* Send user-input command */
            %%send_command(ssId, pVar, pVar->UserInStr);
            strncpy(DetInStr, buffer, sizeof(DetInStr));
            pvPut(DetInStr);
        } state monitor_changes

        when(efTestAndClear(SetTemp_mon)) {
            /* Send set temperature command */
            snprintf(buffer, sizeof(buffer), "set_temperature,%f", SetTemp);
            %%send_command(ssId, pVar, pVar->buffer);
        } state monitor_changes

        when(efTestAndClear(NumFrames_mon)) {
            /* Send number of frames command */
            snprintf(buffer, sizeof(buffer), "set_nframes,%d", NumFrames);
            %%send_command(ssId, pVar, pVar->buffer);
        } state monitor_changes

        when(efTestAndClear(FilePath_mon)) {
            /* Make sure path has a trailing \ */
/* THIS IS BROKEN DUE TO BUG IN SNC 
            if ((FilePath[strlen(FilePath)-1] != '\\') &&
                (strlen(FilePath) < sizeof(FilePath)-2)) {
               strcat(FilePath, "\\");
               pvPut(FilePath);
            }
*/
        } state monitor_changes

        when(efTestAndClear(SaveFile_mon) && 
                            SaveFile==1) {
            /* Set the PV back to 0 */
            SaveFile = 0; pvPut(SaveFile);
            /* Save file */
            save_single = 1;
        } state save_file

        when((efTestAndClear(AcquireCLBK_mon) && AcquireCLBK==1) ||
             (efTestAndClear(AcquirePOLL_mon) && AcquirePOLL==1)) {
            /* Start acquisition */
            exposureTime = Hours*3600. + Minutes*60. + Seconds + Milliseconds/1000.;
            totalTime = exposureTime;
            if (ShutterMode == SHUTTER_EPICS) {
               totalTime += OpenShutterDly + CloseShutterDly;
            }
            snprintf(buffer, sizeof(buffer), "set_exposure,%f", totalTime);
            %%send_command(ssId, pVar, pVar->buffer);
            DetectorState = DETECTOR_ACQUIRE; pvPut(DetectorState);
            /* Set the the start time for the TimeRemaining counter */
            %%epicsTimeGetCurrent(&startTime);
            TimeRemaining = exposureTime; pvPut(TimeRemaining);
            if (FrameType != FRAME_NORMAL) {
                /* Background and flat field save file immediately, so we need to build file name here */
                snprintf(FullFilename, sizeof(FullFilename), FilenameFormat, FileTemplate, SeqNumber);
                pvPut(FullFilename);
                SeqNumber++; pvPut(SeqNumber);
            }
            if (FrameType == FRAME_NORMAL) 
                snprintf(buffer, sizeof(buffer), "start,%d,%d", CorrectBackground, CorrectFlatfield);
            else if (FrameType == FRAME_BACKGROUND)
                snprintf(buffer, sizeof(buffer), "acquire_background,%s%s", FilePath, FullFilename);
            else if (FrameType == FRAME_FLATFIELD) 
                snprintf(buffer, sizeof(buffer), "acquire_flatfield,%s%s", FilePath, FullFilename);
            %%send_command(ssId, pVar, pVar->buffer);
            if (ShutterMode == SHUTTER_EPICS) {
                epicsThreadSleep(OpenShutterDly); 
                strncpy(Shutter, OpenShutterStr, sizeof(Shutter)); pvPut(Shutter);
                /* Wait for the exposure time using epicsEventWaitWithTimeout, so we can abort */
                %%epicsEventWaitWithTimeout(abortEvent, pVar->exposureTime);
                /* Close shutter */
                strncpy(Shutter, CloseShutterStr, sizeof(Shutter)); pvPut(Shutter);
                epicsThreadSleep(CloseShutterDly);
            }
        } state wait_readout
    }

    state wait_readout {
        when (DetectorState != DETECTOR_IDLE && 
              delay(0.05)) {
            /* Read the status */
            %%send_command(ssId, pVar, "get_state");
            DetectorState = atoi(buffer); pvPut(DetectorState);
        } state wait_readout

        when (DetectorState == DETECTOR_IDLE) {
            if (ComputeROICts) {
               %%send_command(ssId, pVar, "get_roi_counts");
               sscanf(buffer, "%lf,%lf", &ROITotal, &ROINet);
               pvPut(ROITotal); pvPut(ROINet);
            }
        } state save_file
    }

    state save_file {
        when() {
            if ((AutoSave && (FrameType == FRAME_NORMAL)) || save_single) {
                /* Send command commands */
                snprintf(buffer, sizeof(buffer), "set_comment,1,%s", Comment1);
                %%send_command(ssId, pVar, pVar->buffer);
                snprintf(buffer, sizeof(buffer), "set_comment,2,%s", Comment2);
                %%send_command(ssId, pVar, pVar->buffer);
                snprintf(buffer, sizeof(buffer), "set_comment,3,%s", Comment3);
                %%send_command(ssId, pVar, pVar->buffer);
                snprintf(buffer, sizeof(buffer), "set_comment,4,%s", Comment4);
                %%send_command(ssId, pVar, pVar->buffer);
                snprintf(buffer, sizeof(buffer), "set_comment,5,%s", Comment5);
                %%send_command(ssId, pVar, pVar->buffer);
                /* Save file */
                snprintf(FullFilename, sizeof(FullFilename), FilenameFormat, FileTemplate, SeqNumber);
                pvPut(FullFilename);
                SeqNumber++; pvPut(SeqNumber);
                snprintf(buffer, sizeof(buffer), "save,%s%s", FilePath, FullFilename);
                %%send_command(ssId, pVar, pVar->buffer);
                save_single = 0;
            }
            /* Set PVs back to 0 */
            AcquireCLBK = 0; pvPut(AcquireCLBK);
            AcquirePOLL = 0; pvPut(AcquirePOLL);
        } state monitor_changes
    }
}


/* This is a separate state set so the watchdog runs even during long acquisitions */
ss watchdog {
    state poll {
        when(delay(1.0) && SNLWatchdog < 5) {
            SNLWatchdog = 5; pvPut(SNLWatchdog);
        } state poll
    }
}

/* This is a separate state set so we can abort during the epicsEventWaitWithTimeout that times the shutter */
ss abort {
    state poll {
        when(Abort_mon && Abort==1) {
            %%send_command(ssId, pVar, "abort");
            %%epicsEventSignal(abortEvent);
            Abort=0; pvPut(Abort);
            TimeRemaining=0.; pvPut(TimeRemaining);
        } state poll
    }
}

/* This is a separate state set so we update the time remaining during the epicsEventWaitWithTimeout that times the shutter */
ss timeRemaining {
    state poll {
        when(delay(.1) && TimeRemaining != 0.) {
            %%epicsTimeGetCurrent(&currentTime);
            %%pVar->TimeRemaining = pVar->exposureTime - epicsTimeDiffInSeconds(&currentTime, &startTime);
            if (TimeRemaining < 0.) TimeRemaining = 0.;
            pvPut(TimeRemaining);
        } state poll
    }
}

/* Begin C code functions */
%{

static int send_command(SS_ID ssId, struct UserVar *pVar, char *command)
{
   int status, len;
   char buffer[MAX_COMMAND_SIZE];
   epicsTimeStamp ts;
   char timeBuff[MAX_COMMAND_SIZE];

   if (pVar->ConnectState != STATE_CONNECTED) return(-1);
   snprintf(buffer, sizeof(buffer), "%s\n", command);
   status = write_socket(pVar->socket, buffer, strlen(buffer)+1);
   strncpy(pVar->DetOutStr, command, sizeof(pVar->DetOutStr));
   seq_pvPut(ssId, pVar->DetOutStrIndex, 0);
   if (pVar->DebugFlag) {
      epicsTimeGetCurrent(&ts);
      epicsTimeToStrftime(timeBuff, sizeof(timeBuff), "%Y/%m/%d %H:%M:%S.%03f", &ts);
      errlogPrintf("%s roperCCD:sent %s\n", timeBuff, command);
   }
   if (status < 0) pVar->ConnectState = STATE_DISCONNECTED;
   /* Should really do a pvPut of ConnectState if status != 0 */

   /* All commands send a response.  Read it into pVar->buffer */
   if (pVar->ConnectState != STATE_CONNECTED) return(-1);
   strcpy(pVar->buffer, "");
   status = listen_socket(pVar->buffer, sizeof(pVar->buffer), 1, 0, 0);
   if (status <= 0) strcpy(pVar->buffer, "");
   /* Remove CR/LF */
   len = strlen(pVar->buffer);
   if (len >= 2) {
      pVar->buffer[len-1]=0;
      pVar->buffer[len-2]=0;
   }
   strncpy(pVar->DetInStr, pVar->buffer, sizeof(pVar->DetInStr));
   seq_pvPut(ssId, pVar->DetInStrIndex, 0);
   if (pVar->DebugFlag) {
      epicsTimeGetCurrent(&ts);
      epicsTimeToStrftime(timeBuff, sizeof(timeBuff), "%Y/%m/%d %H:%M:%S.%03f", &ts);
      errlogPrintf("%s roperCCD:receive (%d) %s\n", timeBuff, status, pVar->buffer);
   }
   return(status);
}

}%
